﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Xml;
using HaloMap;

namespace Gravemind
{
    public partial class ChunkClonner : Form
    {
        public Map map;
        public MapForm form;

        public ChunkClonner(MapForm form, Map map)
        {
            InitializeComponent();
            this.map = map;
            this.form = form;
        }

        private void ChunkClonner_Load(object sender, EventArgs e)
        {
            LoadReflexives(map);
            treeView1.Nodes.Add(map.SelectedTag.Path);
            foreach (Reflexive r in map.SelectedTag.Reflexives)
            {
                treeView1.Nodes[0].Nodes.Add(r.Description);
                AddNodes(r, treeView1.Nodes[0].Nodes[treeView1.Nodes[0].Nodes.Count - 1]);
                treeView1.Nodes[0].Nodes[treeView1.Nodes[0].Nodes.Count - 1].Tag = r.Offset;
            }
            treeView1.Nodes[0].Expand();
        }

        private void AddNodes(Reflexive r, TreeNode n)
        {
            if (r.Chunks.Count != 0)
            {
                foreach (Reflexive rr in r.Chunks)
                {
                    n.Nodes.Add(rr.Description);
                    AddNodes(rr, n.Nodes[n.Nodes.Count - 1]);
                    n.Nodes[n.Nodes.Count - 1].Tag = r.Offset;
                }
            }
            else
            {
                for (int i = 0; i < r.ChunkCount; i++)
                {
                    n.Nodes.Add("Chunk #" + i.ToString());
                    n.Nodes[n.Nodes.Count - 1].Tag = -1;
                }
            }
        }

        public static void LoadReflexives(Map map)
        {
            string C = map.SelectedTag.Class.Replace('*', '!');
            C = C.Replace('/', '_');
            C = C.Replace('\\', '_');
            C = C.Replace('<', '_');
            C = C.Replace('>', '_');
            string Location = Application.StartupPath + "\\Plugins\\" + C + ".Xml";

            if (!File.Exists(Location))
            {
                MessageBox.Show("Can not find plugin!");
                return;
            }
            else
            {
                XmlTextReader xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));

                map.SelectedTag.Reflexives.Clear();

                //Read
                while (xr.Read())
                {
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "reflexive":
                                        {
                                            Reflexive r = new Reflexive();
                                            r.Description = xr.GetAttribute("name");
                                            r.Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                            r.ChunkSize = Convert.ToInt32(xr.GetAttribute("size"));
                                            r.Read(map);
                                            map.SelectedTag.Reflexives.Add(r);
                                            if (r.ChunkCount > 0)
                                            {
                                                RecrusivlyLoadReflexives(map, r, xr.ReadSubtree());
                                            }
                                            else
                                            {
                                                xr.Skip();
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                }
                xr.Close();
            }
        }
        private static void RecrusivlyLoadReflexives(Map map, Reflexive r, XmlReader xr)
        {
            xr.ReadStartElement();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            switch (xr.Name.ToLower())
                            {
                                case "reflexive":
                                    {
                                        string Description = xr.GetAttribute("name");
                                        int Offset = Convert.ToInt32(xr.GetAttribute("offset"));
                                        int Size = Convert.ToInt32(xr.GetAttribute("size"));
                                        for (int i = 0; i < r.ChunkCount; i++)
                                        {
                                            Reflexive rr = new Reflexive();
                                            rr.Description = Description + "[" + i.ToString() + "]";
                                            rr.Offset = Offset + (r.Translation - map.SelectedTag.MetaOffset) + (r.ChunkSize * i);
                                            rr.ChunkSize = Size;
                                            rr.Read(map, rr.Offset);
                                            map.SelectedTag.Reflexives[map.SelectedTag.Reflexives.Count - 1].Chunks.Add(rr);
                                            if (xr.NodeType == XmlNodeType.Element)
                                            {
                                                RecrusivlyLoadReflexives(map, rr, xr.ReadSubtree());
                                            }
                                        }
                                        break;
                                    }
                            }
                            break;
                        }
                    case XmlNodeType.EndElement:
                        {
                            break;
                        }
                }
            }
        }

        public ClipboardContainer GetData()
        {
            PluginReader.LoadReflexives(map);
            ClipboardContainer cc = new ClipboardContainer();
            //Chunk
            if ((int)treeView1.SelectedNode.Tag == -1)
            {
                foreach (Reflexive r in map.SelectedTag.Reflexives)
                {
                    if ((int)treeView1.SelectedNode.Parent.Tag == r.Offset)
                    {
                        int Index = Convert.ToInt32(treeView1.SelectedNode.Text.Substring(7, 1));
                        map.br.BaseStream.Position = r.Translation + (Index * r.ChunkSize);
                        cc.Data = map.br.ReadBytes(r.ChunkSize);
                        cc.IsChunk = true;
                        cc.Name = r.Description;
                        return cc;
                    }
                }
            }
            //Reflexive
            foreach (Reflexive r in map.SelectedTag.Reflexives)
            {
                if (r.Offset == (int)treeView1.SelectedNode.Tag)
                {
                    map.br.BaseStream.Position = r.Translation;
                    cc.Data = map.br.ReadBytes(r.ChunkCount * r.ChunkSize);
                    cc.IsChunk = false;
                    cc.Name = r.Description;
                    return cc;
                }
            }
            MessageBox.Show("Invalid Section!");
            cc.Name = "ERROR";
            return cc;
        }

        private void btnCopy_Click(object sender, EventArgs e)
        {
            ClipboardContainer cc = GetData();
            if (cc.Name != "ERROR")
            {
                Clipboard.SetData("HaloData", cc);
                MessageBox.Show("Copied!");
            }
        }

        public void FindNode(TreeNode n, ref int Index)
        {
            if (n.Text != map.SelectedTag.Path)
            {
                Index++;
                FindNode(n.Parent, ref Index);
            }
        }

        public void ShiftReflexive(Reflexive r, int Shift)
        {
            for (int i = 0; i < r.Chunks.Count; i++)
            {
                map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + (r.ChunkSize * i) + 4 + Shift;
                map.bw.Write(r.Chunks[i].RawTranslation + Shift);
                ShiftReflexive(r.Chunks[i], Shift);
            }
        }

        private void btnCloneChunks_Click(object sender, EventArgs e)
        {
            //Chunk
            PluginReader.LoadReflexives(map);
            if ((int)treeView1.SelectedNode.Tag == -1)
            {
                foreach (Reflexive r in map.SelectedTag.Reflexives)
                {
                    //Find the Reflexive
                    if ((int)treeView1.SelectedNode.Parent.Tag == r.Offset)
                    {
                        //Get the Chunk#
                        int Index = Convert.ToInt32(treeView1.SelectedNode.Text.Substring(7, 1));
                        int Count = Convert.ToInt32(txtChunkstoCopy.Text);
                        //Write new Chunk Count
                        map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + r.Offset;
                        map.bw.Write(r.ChunkCount + Count);
                        //Get the chunk data
                        map.br.BaseStream.Position = r.Translation + (Index * r.ChunkSize);
                        byte[] Data = map.br.ReadBytes(r.ChunkSize);
                        //Goto the end of the Data, store EOF
                        map.br.BaseStream.Position = r.Translation + (r.ChunkCount * r.ChunkSize);
                        byte[] EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                        map.bw.BaseStream.Position = r.Translation + (r.ChunkCount * r.ChunkSize);
                        //Write new Data
                        for (int i = 0; i < Count; i++)
                        {
                            map.bw.Write(Data);
                        }
                        map.bw.Write(EOF);
                        int Shift = Data.Length * Count;

                        //Fix the rest of the reflexives for this tag
                        int ReflexIndex = map.SelectedTag.Reflexives.IndexOf(r);
                        for (int i = ReflexIndex + 1; i < map.SelectedTag.Reflexives.Count; i++)
                        {
                            map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + map.SelectedTag.Reflexives[i].Offset + 4;
                            map.bw.Write(map.SelectedTag.Reflexives[i].RawTranslation + Shift);
                        }

                        //Fix the rest of the reflexives in the map
                        int TagIndex = map.FindIndex(map.SelectedTag.Identifier);
                        Tag t = map.SelectedTag;
                        for (int i = TagIndex + 1; i < map.Tags.Count; i++)
                        {
                            map.SelectedTag = map.Tags[i];
                            PluginReader.ParseReflexives(map, Shift);
                            for (int x = 0; x < map.SelectedTag.Reflexives.Count; x++)
                            {
                                map.bw.BaseStream.Position = map.SelectedTag.MetaOffset + map.SelectedTag.Reflexives[x].Offset + 4;
                                map.bw.Write(map.SelectedTag.Reflexives[x].RawTranslation + Shift);
                            }
                            //map.SelectedTag.MetaOffset += Shift;
                            //Check for shared reflexives
                            switch (map.SelectedTag.Class)
                            {
                                //
                            }
                        }

                        //Fix the Meta Offsets
                        map.bw.BaseStream.Position = map.Tags[TagIndex + 1].IndexOffset + Shift;
                        for (int i = TagIndex + 1; i < map.Index.TagCount; i++)
                        {
                            map.bw.BaseStream.Position += 8;
                            map.bw.Write((map.Tags[i].RawMetaOffset += Shift));
                            map.Tags[i].MetaOffset = map.Tags[i].RawMetaOffset - map.Index.SecondaryMagic;
                            map.bw.BaseStream.Position += 4;
                        }

                        MessageBox.Show("Done!");
                    }
                }
            }
            //Reflexive
            else
            {
                MessageBox.Show("Clone chunks not reflexives!");
                return;
            }
        }
    }
}
