﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using ComponentAce.Compression.Libs.zlib;
using HaloMap;

namespace Gravemind
{
    public partial class Main : Form
    {
        int Count = 0;

        public Main()
        {
            InitializeComponent();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Halo 2 Vista Map (*.map)|*.map";
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                MapForm map = new MapForm(ofd.FileName);
                map.MdiParent = this;
                map.Show();
            }
        }

        private void closeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                this.ActiveMdiChild.Close();
            }
            catch (Exception) { }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            MessageBox.Show("Gravemind by Grimdoomer\nThanks go out to:\nNeodos: Finding countless bugs, converting all plugins,\nmaking the splash, and insisting there where bugs when I disagreed\n\nKiwidoggie: Beta testing, and support.");
        }

        private void optionsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Settings s = new Settings();
            s.MdiParent = this;
            s.Show();
        }

        private void resignToolStripMenuItem_Click(object sender, EventArgs e)
        {
            try
            {
                MapForm m = (MapForm)this.ActiveMdiChild;
                m.Resign();
            }
            catch (Exception) { return; }
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (Count == 25)
            {
                this.BackgroundImage = null;
                timer1.Enabled = false;
            }
            else
            {
                Count++;
            }
        }

        private void Main_Load(object sender, EventArgs e)
        {
            //Intialize
            Globals.Initialize();

            //Check for updates
            if (Globals.CheckForUpdates)
            {
                Gravemind.Update.CheckUpdate(this);
            }
        }

        private void tileHorizontalToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayoutMdi(MdiLayout.TileHorizontal);
        }

        private void tileVerticalToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayoutMdi(MdiLayout.TileVertical);
        }

        private void cascadeToolStripMenuItem_Click(object sender, EventArgs e)
        {
            LayoutMdi(MdiLayout.Cascade);
        }

        private void makeDefaultEditorToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (MessageBox.Show("Are you sure you would like to set Gravemind as your default editor?", "Set As Default Editor", MessageBoxButtons.YesNo) == DialogResult.Yes)
            {
                Microsoft.Win32.RegistryKey regKey;
                regKey = Microsoft.Win32.Registry.ClassesRoot.CreateSubKey("Halo 2 Vista Map\\");
                regKey.SetValue("", "Halo 2 Vista Map");
                regKey = Microsoft.Win32.Registry.ClassesRoot.CreateSubKey("Halo 2 Vista Map\\shell\\open\\command");
                regKey.SetValue("", Application.ExecutablePath + " %1");
                regKey = Microsoft.Win32.Registry.ClassesRoot.CreateSubKey(".map");
                regKey.SetValue("", "Halo 2 Vista Map");
                regKey.Close();
                MessageBox.Show("Gravemind is now your default map editor.", "Set Default Editor");
            }
        }

        private void forThisMapToolStripMenuItem_Click(object sender, EventArgs e)
        {
            FolderBrowserDialog ofd = new FolderBrowserDialog();
            ofd.Description = "Please browse for the maps folder on your Halo 2 disk";
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                SaveFileDialog sfd = new SaveFileDialog();
                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    MapForm mform = (MapForm)this.ActiveMdiChild;
                    string map = ofd.SelectedPath + "\\" + mform.Map.Header.MapName + ".map.dtz";

                    //Decompress
                    MemoryStream CompressedFile = new MemoryStream(File.ReadAllBytes(map.Replace("\0", "")));
                    MemoryStream DecompressedFile = new MemoryStream();
                    ZOutputStream ZOut = new ZOutputStream(DecompressedFile);
                    CopyStream(CompressedFile, ZOut);

                    File.WriteAllBytes(sfd.FileName + ".map", DecompressedFile.ToArray());
                    MessageBox.Show("Done!");
                }
            }
        }

        public static void CopyStream(Stream input, Stream output)
        {
            byte[] buffer = new byte[2000];
            int len;
            while ((len = input.Read(buffer, 0, 2000)) > 0)
            {
                output.Write(buffer, 0, len);
            }
            output.Flush();
        }

        private void decompressCacheMapsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog ofd = new OpenFileDialog();
            ofd.Filter = "Halo 2 Cache Map (*.map.dtz)|*.map.dtz";
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                SaveFileDialog sfd = new SaveFileDialog();
                if (sfd.ShowDialog() == DialogResult.OK)
                {
                    MapForm mform = (MapForm)this.ActiveMdiChild;
                    string map = ofd.FileName;

                    //Decompress
                    MemoryStream CompressedFile = new MemoryStream(File.ReadAllBytes(map.Replace("\0", "")));
                    MemoryStream DecompressedFile = new MemoryStream();
                    ZOutputStream ZOut = new ZOutputStream(DecompressedFile);
                    CopyStream(CompressedFile, ZOut);

                    File.WriteAllBytes(sfd.FileName + ".map", DecompressedFile.ToArray());
                    MessageBox.Show("Done!");
                }
            }
        }
    }
}
