﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Microsoft.VisualBasic;
using System.IO;
using System.Xml;
using HaloMap.BSP;

namespace HaloMap
{
    public class Tag : ICloneable
    {
        public string Class = "";
        public string Path;
        public int FileTableOffset;
        public int IndexOffset;
        public int Identifier;
        public int MetaOffset;
        public int RawMetaOffset;
        public int MetaSize;

        public List<TagReference> Idents = new List<TagReference>();
        public List<ScriptReference> ScriptReferences = new List<ScriptReference>();
        public List<Reflexive> Reflexives = new List<Reflexive>();

        #region ICloneable Members

        public object Clone()
        {
            Tag t = new Tag();
            t.Class = (string)Class.Clone();
            t.Path = (string)Path.Clone();
            t.FileTableOffset = FileTableOffset;
            t.IndexOffset = IndexOffset;
            t.Identifier = Identifier;
            t.MetaOffset = MetaOffset;
            t.RawMetaOffset = RawMetaOffset;
            t.MetaSize = MetaSize;
            return t;
        }

        #endregion
    }

    public class Tags
    {
        public void Read(Map map)
        {
            map.Tags.Clear();
            map.br.BaseStream.Position = map.Index.TagStart;

            //Index
            for (int i = 0; i < map.Index.TagCount; i++)
            {
                Tag Tag = new Tag();
                Tag.IndexOffset = (int)map.br.BaseStream.Position;
                Tag.Class = Strings.StrReverse(new string(map.br.ReadChars(4)));
                Tag.Identifier = map.br.ReadInt32();
                Tag.RawMetaOffset = map.br.ReadInt32();
                Tag.MetaOffset = Tag.RawMetaOffset - map.Index.SecondaryMagic;
                Tag.MetaSize = map.br.ReadInt32();
                map.Tags.Add(Tag);
            }

            //FileTable
            map.br.BaseStream.Position = map.Header.FileTableOffset;
            for (int i = 0; i < map.Index.TagCount; i++)
            {
                map.Tags[i].FileTableOffset = (int)map.br.BaseStream.Position;
                char chr;
                while ((chr = map.br.ReadChar()) != '\0')
                    map.Tags[i].Path += chr;
            }
        }

        public void Read(Map map, TreeView TV)
        {
            map.br.BaseStream.Position = map.Index.TagStart;

            //Index
            for (int i = 0; i < map.Index.TagCount; i++)
            {
                Tag Tag = new Tag();
                Tag.IndexOffset = (int)map.br.BaseStream.Position;
                Tag.Class = Strings.StrReverse(new string(map.br.ReadChars(4)));
                Tag.Identifier = map.br.ReadInt32();
                Tag.RawMetaOffset = map.br.ReadInt32();
                Tag.MetaOffset = Tag.RawMetaOffset - map.Index.SecondaryMagic;
                Tag.MetaSize = map.br.ReadInt32();
                map.Tags.Add(Tag);
            }
            
            //FileTable
            map.br.BaseStream.Position = map.Header.FileTableOffset;
            for (int i = 0; i < map.Index.TagCount; i++)
            {
                map.Tags[i].FileTableOffset = (int)map.br.BaseStream.Position;
                char chr;
                while ((chr = map.br.ReadChar()) != '\0')
                    map.Tags[i].Path += chr;
            }

            //TreeView
            TV.Nodes.Add(map.Header.MapName);
            ListBox templistbox = new ListBox();

            for (int i = 0; i < map.Index.TagCount; i++)
            {
                if (templistbox.Items.Contains(map.Tags[i].Class) == false)
                {
                    templistbox.Items.Add(map.Tags[i].Class);
                    TreeNode ClassNode = new TreeNode(map.Tags[i].Class);
                    foreach (Tag t in map.Tags)
                    {
                        if (t.Class == ClassNode.Text)
                        {
                            ClassNode.Nodes.Add(t.Path);
                        }
                    }
                    TV.Nodes[0].Nodes.Add(ClassNode);
                    Application.DoEvents();
                }
            }
            TV.Sort();
            TV.Nodes[0].Expand();
        }

        public void ReadMP(Map map, TreeView TV)
        {
            map.br.BaseStream.Position = map.Index.TagStart;

            //Index
            for (int i = 0; i < map.Index.TagCount; i++)
            {
                Tag Tag = new Tag();
                Tag.IndexOffset = (int)map.br.BaseStream.Position;
                Tag.Class = Strings.StrReverse(new string(map.br.ReadChars(4)));
                Tag.Identifier = map.br.ReadInt32();
                Tag.RawMetaOffset = map.br.ReadInt32();
                Tag.MetaOffset = Tag.RawMetaOffset - map.Index.SecondaryMagic;
                Tag.MetaSize = map.br.ReadInt32();
                if (Tag.MetaSize > 0)
                {
                    map.Tags.Add(Tag);
                }
            }

            //FileTable
            map.br.BaseStream.Position = map.Header.FileTableOffset;
            for (int i = 0; i < map.Tags.Count; i++)
            {
                map.Tags[i].FileTableOffset = (int)map.br.BaseStream.Position;
                char chr;
                while ((chr = map.br.ReadChar()) != '\0')
                    map.Tags[i].Path += chr;
            }

            //TreeView
            TV.Nodes.Add(map.Header.MapName);
            ListBox templistbox = new ListBox();

            for (int i = 0; i < map.Tags.Count; i++)
            {
                if (templistbox.Items.Contains(map.Tags[i].Class) == false)
                {
                    templistbox.Items.Add(map.Tags[i].Class);
                    TreeNode ClassNode = new TreeNode(map.Tags[i].Class);
                    foreach (Tag t in map.Tags)
                    {
                        if (t.Class == ClassNode.Text)
                        {
                            ClassNode.Nodes.Add(t.Path);
                        }
                    }
                    TV.Nodes[0].Nodes.Add(ClassNode);
                    Application.DoEvents();
                }
            }
            TV.Sort();
            TV.Nodes[0].Expand();
        }

        #region Filters
        public static List<Tag> FilterSounds(Map map)
        {
            List<Tag> Sounds = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "snd!")
                {
                    Sounds.Add(t);
                }
            }
            return Sounds;
        }
        public static List<Tag> FilterModels(Map map)
        {
            List<Tag> Models = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "mode")
                {
                    Models.Add(t);
                }
            }
            return Models;
        }
        public static List<Tag> FilterBSP(Map map)
        {
            List<Tag> BSPs = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "SBSP")
                {
                    BSPs.Add(t);
                }
            }
            return BSPs;
        }
        public static List<Tag> FilterLightmaps(Map map)
        {
            List<Tag> Lightmaps = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "ltmp")
                {
                    Lightmaps.Add(t);
                }
            }
            return Lightmaps;
        }
        public static List<Tag> FilterDecorators(Map map)
        {
            List<Tag> Decorators = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "DECR")
                {
                    Decorators.Add(t);
                }
            }
            return Decorators;
        }
        public static List<Tag> FilterWeather(Map map)
        {
            List<Tag> Weather = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "weat")
                {
                    Weather.Add(t);
                }
            }
            return Weather;
        }
        public static List<Tag> FilterParticles(Map map)
        {
            List<Tag> Particles = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "PRTM")
                {
                    Particles.Add(t);
                }
            }
            return Particles;
        }
        public static List<Tag> FilterCollision(Map map)
        {
            List<Tag> Collision = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "coll")
                {
                    Collision.Add(t);
                }
            }
            return Collision;
        }
        public static List<Tag> FilterAnimations(Map map)
        {
            List<Tag> Animations = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "jmad")
                {
                    Animations.Add(t);
                }
            }
            return Animations;
        }
        public static List<Tag> FilterBitmaps(Map map)
        {
            List<Tag> Bitmaps = new List<Tag>();
            foreach (Tag t in map.Tags)
            {
                if (t.Class == "bitm")
                {
                    Bitmaps.Add(t);
                }
            }
            return Bitmaps;
        }
        #endregion

        public void ExtractMeta(Map map, Tag Tag, string Location)
        {
            //Load all the references
            Tag OldTag = map.SelectedTag;
            map.SelectedTag = Tag;
            PluginReader.ReadReferences(map);

            string Class = map.SelectedTag.Class.Replace('*', '!');
            Class = Class.Replace('/', '_');
            Class = Class.Replace('\\', '_');
            Class = Class.Replace('<', '_');
            Class = Class.Replace('>', '_');

            //Write Xml
            XmlWriter xr = XmlWriter.Create(Location + "." + Class);
            xr.WriteStartDocument();
            xr.WriteStartElement("Tag");
            xr.WriteElementString("Class", Tag.Class);
            xr.WriteElementString("Path", Tag.Path);
            xr.WriteElementString("Ident", Tag.Identifier.ToString());
            xr.WriteElementString("Size", Tag.MetaSize.ToString());
            xr.WriteElementString("IdentCount", Tag.Idents.Count.ToString());
            xr.WriteElementString("ReflexiveCount", Tag.Reflexives.Count.ToString());
            xr.WriteElementString("ScriptCount", Tag.ScriptReferences.Count.ToString());

            //Write Tag Refs
            foreach (TagReference t in map.SelectedTag.Idents)
            {
                xr.WriteStartElement("TagReference");
                xr.WriteElementString("Class", t.Class);
                xr.WriteElementString("Path", t.Tag.Path);
                xr.WriteElementString("Ident", t.Tag.Identifier.ToString());
                xr.WriteElementString("Offset", t.Offset.ToString());
                xr.WriteEndElement();
            }

            //Write Reflexives
            foreach (Reflexive r in map.SelectedTag.Reflexives)
            {
                xr.WriteStartElement("Reflexive");
                xr.WriteElementString("ChunkCount", r.ChunkCount.ToString());
                xr.WriteElementString("ChunkSize", r.ChunkSize.ToString());
                xr.WriteElementString("Translation", (r.Translation - map.SelectedTag.MetaOffset).ToString());
                xr.WriteElementString("Offset", r.Offset.ToString());
                xr.WriteEndElement();
            }

            //Write String Refs
            foreach (ScriptReference s in map.SelectedTag.ScriptReferences)
            {
                xr.WriteStartElement("ScriptReference");
                xr.WriteElementString("Offset", s.Offset.ToString());
                xr.WriteElementString("Index", s.Index.ToString());
                xr.WriteElementString("Reference", s.Reference);
                xr.WriteEndElement();
            }

            xr.WriteEndElement();
            xr.Close();

            //Write Meta
            BinaryWriter bw = new BinaryWriter(new FileStream(Location, FileMode.OpenOrCreate, FileAccess.Write, FileShare.Write));
            map.br.BaseStream.Position = Tag.MetaOffset;
            bw.Write(map.br.ReadBytes(Tag.MetaSize));
            bw.Close();

            map.SelectedTag = OldTag;
        }

        public void MorphMeta(Map map, Tag tag, string Location)
        {
            Tag oldtag = map.SelectedTag;
            map.SelectedTag = tag;
            PluginReader.ReadReferences(map);

            map.br.BaseStream.Position = tag.MetaOffset;
            if (File.Exists(Location + "." + tag.Class))
            {
                return;
            }
            string Class = map.SelectedTag.Class.Replace('*', '!');
            Class = Class.Replace('/', '_');
            Class = Class.Replace('\\', '_');
            Class = Class.Replace('<', '_');
            Class = Class.Replace('>', '_');
            BinaryWriter bw = new BinaryWriter(new FileStream(Location + "." + Class, FileMode.Create, FileAccess.Write, FileShare.Write));
            bw.Write(map.br.ReadBytes(tag.MetaSize));

            //Change Idents
            foreach (TagReference t in map.SelectedTag.Idents)
            {
                try
                {
                    bw.BaseStream.Position = t.Offset + 4;
                    bw.Write(map.FindIndex(t.Ident));
                }
                catch { }
            }

            //Change Reflexives
            foreach (Reflexive r in map.SelectedTag.Reflexives)
            {
                try
                {
                    bw.BaseStream.Position = r.Offset + 4;
                    bw.Write(r.Translation - tag.MetaOffset);
                }
                catch { }
            }

            //Change Strings
            foreach (ScriptReference s in map.SelectedTag.ScriptReferences)
            {
                try
                {
                    bw.BaseStream.Position = s.Offset;
                    bw.Write(map.FindScript(s.Reference));
                }
                catch { }
            }

            bw.Close();
            map.SelectedTag = oldtag;
        }

        public void InjectMeta(Map map, Tag Tag, string Location)
        {
            XmlTextReader xr = new XmlTextReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));
            int MetaSize = 0;
            Tag.Idents.Clear();
            Tag.Reflexives.Clear();
            Tag.ScriptReferences.Clear();
            while (xr.Read())
            {
                switch (xr.NodeType)
                {
                    case XmlNodeType.Element:
                        {
                            if (xr.Name != "Tag")
                            {
                                MessageBox.Show("Invalid Meta!");
                                return;
                            }

                            xr.ReadStartElement();
                            xr.ReadElementString("Class");
                            xr.ReadElementString("Path");
                            xr.ReadElementString("Ident");
                            MetaSize = Convert.ToInt32(xr.ReadElementString("Size"));
                            if (MetaSize > Tag.MetaSize)
                            {
                                MessageBox.Show("New Meta is to big!");
                                return;
                            }
                            int IdentCount = Convert.ToInt32(xr.ReadElementString("IdentCount"));
                            int ReflexiveCount = Convert.ToInt32(xr.ReadElementString("ReflexiveCount"));
                            int ScriptCount = Convert.ToInt32(xr.ReadElementString("ScriptCount"));

                            //Idents
                            for (int i = 0; i < IdentCount; i++)
                            {
                                TagReference t = new TagReference();
                                t.Tag = new Tag();
                                xr.ReadStartElement();
                                t.Tag.Class = xr.ReadElementString("Class");
                                t.Tag.Path = xr.ReadElementString("Path");
                                t.Tag.Identifier = Convert.ToInt32(xr.ReadElementString("Ident"));
                                t.Offset = Convert.ToInt32(xr.ReadElementString("Offset"));
                                xr.ReadEndElement();
                                Tag.Idents.Add(t);
                            }

                            //Reflexives
                            for (int i = 0; i <ReflexiveCount; i++)
                            {
                                Reflexive r = new Reflexive();
                                xr.ReadStartElement();
                                r.ChunkCount = Convert.ToInt32(xr.ReadElementString("ChunkCount"));
                                r.ChunkSize = Convert.ToInt32(xr.ReadElementString("ChunkSize"));                                                
                                r.Translation = Convert.ToInt32(xr.ReadElementString("Translation"));
                                r.Offset = Convert.ToInt32(xr.ReadElementString("Offset"));
                                xr.ReadEndElement();
                                Tag.Reflexives.Add(r);
                            }
                                            
                            //Scripts
                            for (int i = 0; i <ScriptCount; i++)
                            {
                                ScriptReference s = new ScriptReference();
                                xr.ReadStartElement();
                                s.Offset = Convert.ToInt32(xr.ReadElementString("Offset"));
                                s.Index = Convert.ToInt32(xr.ReadElementString("Index"));
                                s.Reference = xr.ReadElementString("Reference");
                                xr.ReadEndElement();
                                Tag.ScriptReferences.Add(s);
                            }
                            break;
                        }
                }
            }

            BinaryReader br = new BinaryReader(new FileStream(Location.Remove(Location.LastIndexOf(".")), FileMode.Open, FileAccess.Read, FileShare.Read));

            map.bw.BaseStream.Position = Tag.MetaOffset;
            map.bw.Write(br.ReadBytes((int)br.BaseStream.Length));
            br.Close();
            for (int i = MetaSize; i < Tag.MetaSize; i++)
            {
                map.bw.Write((byte)0);
            }

            //Relink Idents
            foreach (TagReference t in Tag.Idents)
            {
                map.bw.BaseStream.Position = Tag.MetaOffset + t.Offset;
                t.Tag = map.FindTag(t.Ident);
                t.Write(map, (int)map.bw.BaseStream.Position);
            }

            //Fix Reflexive Translations
            foreach (Reflexive r in Tag.Reflexives)
            {
                map.bw.BaseStream.Position = Tag.MetaOffset + r.Offset;
                r.Translation = (r.Translation + Tag.MetaOffset) + map.Index.SecondaryMagic;
                r.Write(map, (int)map.bw.BaseStream.Position);
            }

            //Relink Script References
            foreach (ScriptReference s in Tag.ScriptReferences)
            {
                map.bw.BaseStream.Position = Tag.MetaOffset + s.Offset;
                s.Index = map.FindScript(s.Reference);
                s.Reference = map.ScriptReferences[s.Index].Reference;
                s.Write(map, (int)map.bw.BaseStream.Position);
            }
        }

        public void RecrusiveExtract(Map map, string Location)
        {
            //Verify Settings
            if (map.VerifySettings() == false)
            {
                MessageBox.Show("Error finding external map!", "Operation Aborted");
                return;
            }

            //Create our Directory
            int TempIndex = map.SelectedTag.Path.LastIndexOf("\\") + 1;

            //Create the File and open it in a StreamReader
            StreamWriter sw = new StreamWriter(new FileStream(Location + "\\" + map.SelectedTag.Path.Substring(TempIndex, map.SelectedTag.Path.Length - TempIndex) + ".Tag", FileMode.Create));

            SaveRecrusive(map, sw, map.SelectedTag, Location);

            sw.Close();
        }

        private void SaveRecrusive(Map map, StreamWriter sw, Tag Tag, string Location)
        {
            //Create Sub-Directorys
            int Pointer = Tag.Path.LastIndexOf("\\") + 1;
            string SubPath = "\\" + Tag.Path.Substring(0, Pointer);
            Directory.CreateDirectory(Location + SubPath);

            //Name of file
            string TempName = Tag.Path.Substring(Pointer, Tag.Path.Length - Pointer);
            string TempTagName = SubPath + TempName; // +"." + Tag.Class;
            TempTagName = TempTagName.Replace('*', '!');
            TempTagName = TempTagName.Replace("<", "_");
            TempTagName = TempTagName.Replace(">", "_");
            //TempTagName = TempTagName.Replace('/', '_');
            //TempTagName = TempTagName.Replace('\\', '_');

            //Save the Tag
            ExtractMeta(map, Tag, Location + TempTagName);

            //Save Raw
            SaveRaw(map, Tag, Location);

            //Recrusivly save idents
            foreach (TagReference t in Tag.Idents)
            {
                if (t.Tag.Identifier != -1)
                {
                    SaveRecrusive(map, sw, t.Tag, Location);

                    //Save Raw Data
                    SaveRaw(map, t.Tag, Location);
                }
            }

            //Add it to the list
            sw.WriteLine(TempTagName + "." + Tag.Class);
        }

        public void SaveRaw(Map map, Tag tag, string Location)
        {
            if (tag.Path != "Null" && tag.Class != "Null")
            {
                switch (tag.Class)
                {
                    case "bitm":
                        {
                            BitmapRaw.Decompile(map, tag, Location + "\\" + tag.Path);
                            break;
                        }
                    case "mode":
                        {
                            ModelRaw.Decompile(map, tag, Location + "\\" + tag.Path);
                            break;
                        }
                }
            }
        }

        public void Build(Map map, string Location)
        {
            //Check the tag and warn the user
            if (map.Header.MapType != MapType.Multiplayer && map.Header.MapType != MapType.Singleplayer)
            {
                if (MessageBox.Show("Warning! This map is an extrenal reference for all other maps,\nBuilding tags to this map could cause other maps not to load.\nContinue?", "Warning External Reference", MessageBoxButtons.YesNo) == DialogResult.No)
                {
                    return;
                }
            }

            List<string> Tags = new List<string>();
            List<int> Idents = new List<int>();

            //Get the Directory
            string Dir = Location.Substring(0, Location.LastIndexOf("\\"));

            //Read the tags in the build file
            StreamReader sr = new StreamReader(new FileStream(Location, FileMode.Open, FileAccess.Read, FileShare.Read));
            while (sr.EndOfStream == false)
            {
                Tags.Add(sr.ReadLine());
                Idents.Add(map.Tags[map.Index.TagCount - 1].Identifier + (65537 * Tags.Count));
            }

            //Get all Raw tags
            List<Tag> BSPs = HaloMap.Tags.FilterBSP(map);
            List<Tag> Lightmaps = HaloMap.Tags.FilterLightmaps(map);
            List<Tag> Models = HaloMap.Tags.FilterModels(map);
            List<Tag> Decorators = HaloMap.Tags.FilterDecorators(map);
            List<Tag> Particles = HaloMap.Tags.FilterParticles(map);
            List<Tag> Weather = HaloMap.Tags.FilterWeather(map);
            List<Tag> Collision = HaloMap.Tags.FilterCollision(map);
            List<Tag> Animations = HaloMap.Tags.FilterAnimations(map);
            List<Tag> Bitmaps = HaloMap.Tags.FilterBitmaps(map);

            //Write each tag
            for (int i = 0; i < Tags.Count; i++)
            {
                #region Setup
                //Update
                map.UpdateProccess("Building (" + i.ToString() + "/" + Tags.Count.ToString() + ")...");

                //Set up our pointers
                int Shift = 0;
                bool New = false;
                int PermIndex = -1;
                int Padding;
                int OldPadding;
                byte[] EOF;
                //Pull out Ugh
                Tag Ugh = map.Tags[map.Index.TagCount - 1];
                RawChunk Raw = new RawChunk();
                RawType RawType;
                #endregion

                #region Sound Raw
                //Write Internel Sound Raw
                if (Tags[i].EndsWith(".snd!"))
                {
                    //Set Type
                    RawType = RawType.Sound;

                    //Goto the Model Raw Pointer - 256
                    map.br.BaseStream.Position = map.Header.ModelRawTableStart;

                    //Open and read raw
                    Raw = new RawChunk();
                    Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                    Raw.Offset = (int)map.br.BaseStream.Position;
                    Raw.Size = Raw.Raw.Length;
                    Raw.RawLocation = RawLocation.Internal;

                    //Store EOF
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                    //Write Raw
                    map.bw.BaseStream.Position = Raw.Offset;
                    map.bw.Write(Raw.Raw);

                    //Write Padding
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);

                    //WriteEOF
                    map.bw.Write(EOF);

                    //Add Shift
                    Shift += Raw.Size;
                    Shift += Padding;

                    //Add to Crazy
                    Reflexive Temp = new Reflexive();

                    #region Read and shift each Reflexive
                    //Read and shift each Reflexive first
                    //--------------------------SoundNames
                    map.br.BaseStream.Position = Ugh.MetaOffset + Shift + 16;
                    Reflexive SoundNames = new Reflexive();
                    SoundNames.Read(map, true);
                    SoundNames.ChunkCount += 1;
                    SoundNames.Write(map, true);

                    //--------------------------Unknown24
                    map.br.BaseStream.Position = Ugh.MetaOffset + Shift + 24;
                    Temp.Read(map, true);
                    Temp.Translation += 4;
                    Temp.Write(map, true);

                    //--------------------------Permutations
                    map.br.BaseStream.Position = Ugh.MetaOffset + 32 + Shift;
                    Reflexive Permutations = new Reflexive();
                    Permutations.Read(map, true);
                    Permutations.Translation += 4;
                    Permutations.ChunkCount += 1;
                    Permutations.Write(map, true);

                    //--------------------------SoundChoices
                    map.br.BaseStream.Position = Ugh.MetaOffset + 40 + Shift;
                    Reflexive SoundChoices = new Reflexive();
                    SoundChoices.Read(map, true);
                    SoundChoices.Translation += 16;
                    SoundChoices.ChunkCount += 1;
                    SoundChoices.Write(map, true);

                    //--------------------------Unknown51
                    map.br.BaseStream.Position = Ugh.MetaOffset + 40 + Shift;
                    Temp.Read(map, true);
                    //Unknown53
                    for (int x = 0; x < Temp.ChunkCount; x++)
                    {
                        map.br.BaseStream.Position = Temp.Translation + Shift + (52 * x) + 20;
                        Reflexive Unknown53 = new Reflexive();
                        Unknown53.Read(map, true);
                        Unknown53.Translation += 32;
                        Unknown53.Write(map, true);
                    }
                    Temp.Translation += 16;
                    Temp.Write(map, true);

                    //--------------------------Zero56
                    map.br.BaseStream.Position = Ugh.MetaOffset + 56 + Shift;
                    Temp.Read(map, true);
                    Temp.Translation += 32;
                    Temp.Write(map, true);

                    //--------------------------SoundChunks
                    map.br.BaseStream.Position = Ugh.MetaOffset + 64 + Shift;
                    Reflexive SoundChunks = new Reflexive();
                    SoundChunks.Read(map, true);
                    SoundChunks.Translation += 32;
                    SoundChunks.ChunkCount += 1;
                    SoundChunks.Write(map, true);

                    //--------------------------Unknown72
                    map.br.BaseStream.Position = Ugh.MetaOffset + 72 + Shift;
                    Temp.Read(map, true);
                    for (int x = 0; x < Temp.ChunkCount; x++)
                    {
                        map.br.BaseStream.Position = Temp.Translation + Shift + (28 * x);
                        Reflexive Unknown74 = new Reflexive();
                        Unknown74.Read(map, true);
                        Unknown74.Translation += 44;
                        Unknown74.Write(map, true);
                        map.br.BaseStream.Position = Temp.Translation + Shift + (28 * x) + 8;
                        Unknown74.Read(map, true);
                        Unknown74.Translation += 44;
                        Unknown74.Write(map, true);
                    }
                    Temp.Translation += 44;
                    Temp.Write(map, true);

                    //--------------------------Model Raw
                    map.br.BaseStream.Position = Ugh.MetaOffset + 80 + Shift;
                    Temp.Read(map, true);
                    for (int x = 0; x < Temp.ChunkCount; x++)
                    {
                        map.br.BaseStream.Position = Temp.Translation + Shift + (44 * x) + 24;
                        Reflexive Unknown74 = new Reflexive();
                        Unknown74.Read(map, true);
                        Unknown74.Translation += 44;
                        Unknown74.Write(map, true);
                    } 
                    Temp.Translation += 44;
                    Temp.Write(map, true);
                    #endregion

                    #region Write new Crazy Meta
                    //-------------Sound Names
                    map.bw.BaseStream.Position = SoundNames.Translation + Shift + (SoundNames.ChunkCount * 4) + 4;
                    map.bw.Write((int)0);

                    //-------------Permutations
                    map.bw.BaseStream.Position = Permutations.Translation + Shift + (Permutations.ChunkCount * 12);
                    map.bw.Write((short)0);
                    map.bw.Write((short)0);
                    map.bw.Write((byte)0);
                    map.bw.Write(Permutations.ChunkCount - 1);
                    map.bw.Write((short)0);
                    map.bw.Write((short)SoundChoices.ChunkCount - 1);
                    map.bw.Write((short)1);

                    //-------------Choices
                    map.bw.BaseStream.Position = SoundChoices.Translation + Shift + (SoundChoices.ChunkCount * 16) + 16;
                    map.bw.Write((int)SoundNames.ChunkCount - 1);
                    map.bw.Write((int)0);
                    map.bw.Write((int)0);
                    map.bw.Write((short)SoundChunks.ChunkCount - 1);
                    map.bw.Write((short)1);

                    //-------------SoundChunks
                    map.bw.BaseStream.Position = SoundChunks.Translation + Shift + (SoundChunks.ChunkCount * 12) + 12;
                    map.bw.Write(Raw.Offset);
                    map.bw.Write(Raw.Size);
                    #endregion

                    //Set Permutation Index
                    PermIndex = Permutations.ChunkCount - 1;

                    //Shift Model Raw Start
                    map.Header.ModelRawTableStart += Shift;
                }
                #endregion

                #region Model Raw
                //Write Internel Model Raw
                if (Tags[i].EndsWith(".mode"))
                {
                    //Set Raw Type
                    RawType = RawType.Model;

                    //Goto the Model Raw Table
                    map.br.BaseStream.Position = map.Header.ModelRawTableStart;

                    //Open and Read Raw
                    Raw = new RawChunk();
                    Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                    Raw.Offset = (int)map.br.BaseStream.Position;
                    Raw.Size = Raw.Raw.Length;
                    Raw.RawLocation = RawLocation.Internal;

                    //Store EOF
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                    //Write Raw
                    map.bw.BaseStream.Position = Raw.Offset;
                    map.bw.Write(Raw.Raw);

                    //Write Padding
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);

                    //Write EOF
                    map.bw.Write(EOF);

                    //Add to Shift
                    Shift += Raw.Size;
                    Shift += Padding;

                    //Shift Raw Table Size
                    map.Header.ModelRawTableSize += Shift;
                }
                #endregion

                #region BSP Raw
                //Write Internal BSP Raw
                if (Tags[i].EndsWith(".sbsp"))
                {
                    //Set Raw Type
                    RawType = RawType.BSP;

                    //Get the BSP index to write to
                    if (BSPs.Count != 8)
                    {
                        if (MessageBox.Show("Would you like to overwrite an existing Bsp(Yes)?\nOr create a new one(No)?", "Bsp Tag Detected", MessageBoxButtons.YesNo) == DialogResult.No)
                        {
                            goto CreateNew;
                        }
                        //OverWrite
                        goto OverWrite;
                    }
                    //OverWrite
                    goto OverWrite;

                CreateNew:
                    {
                        New = true;

                        //We go the first Lightmap raw offset
                        map.br.BaseStream.Position = map.Lightmaps[0].Groups[0].Clustors[0].RawOffset;

                        //Open and Read Raw
                        Raw = new RawChunk();
                        Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                        Raw.Offset = (int)map.br.BaseStream.Position;
                        Raw.Size = Raw.Raw.Length;
                        Raw.RawLocation = RawLocation.Internal;

                        //Store EOF
                        EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                        //Write Raw
                        map.bw.BaseStream.Position = Raw.Offset;
                        map.bw.Write(Raw.Raw);

                        //Write Padding
                        Padding = map.Padding();
                        map.bw.Write(new byte[Padding]);

                        //Add to Shift
                        Shift += Raw.Size;
                        Shift += Padding;

                        //Shift Raw Table Size
                        map.Header.ModelRawTableSize += Shift;
                    }

                OverWrite:
                    {
                        New = false;

                        //We need to get the raw pointers for the 1st Bsp
                    }
                }
                #endregion

                #region Lightmap Raw
                //Write Internal Lightmap Raw
                if (Tags[i].EndsWith(".ltmp"))
                {
                    //Write New
                    if (New == true)
                    {
                        //Goto the last lightmap pointer
                        for (int x = 0; x < map.Lightmaps[map.Lightmaps.Count - 1].Groups[map.Lightmaps[map.Lightmaps.Count - 1].Groups.Count - 1].Clustors.Count; x++)
                        {
                            Clustor c = map.Lightmaps[map.Lightmaps.Count - 1].Groups[map.Lightmaps[map.Lightmaps.Count - 1].Groups.Count - 1].Clustors[x];
                            if (c.RawOffset > 0 && c.RawSize > 0)
                            {
                                map.br.BaseStream.Position = c.RawOffset + c.RawSize;
                            }
                        }

                        //Open and Read Raw
                        Raw = new RawChunk();
                        Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                        Raw.Offset = (int)map.br.BaseStream.Position;
                        Raw.Size = Raw.Raw.Length;
                        Raw.RawLocation = RawLocation.Internal;

                        //Store EOF
                        EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                        //Write Raw
                        map.bw.BaseStream.Position = Raw.Offset;
                        map.bw.Write(Raw.Raw);

                        //Padding
                        Padding = map.Padding();
                        map.bw.Write(new byte[Padding]);

                        //Add to Shift
                        Shift += Raw.Size;
                        Shift += Padding;

                        //Shift Raw Table Size
                        map.Header.ModelRawTableSize += Shift;
                    }
                }
                #endregion

                #region Decorator Raw
                if (Tags[i].EndsWith(".DECR"))
                {
                    //Set Raw Type
                    RawType = RawType.Decorator;

                    //Goto the last Decorator offset
                    int Off = 0;
                    int Size = 0;
                    for (int x = 0; x < Decorators.Count; x++)
                    {
                        map.br.BaseStream.Position = Decorators[x].MetaOffset + 56;
                        int Offset = map.br.ReadInt32();
                        int Size2 = map.br.ReadInt32();
                        if (Offset > 0 && Size > 0 && RawData.RawLocator(Offset) == RawLocation.Internal)
                        {
                            if (Offset > Off)
                            {
                                Off = Offset;
                                Size = Size2;
                            }
                        }
                    }

                    map.br.BaseStream.Position = Off + Size;

                    //Open and read raw
                    Raw = new RawChunk();
                    Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                    Raw.Offset = (int)map.br.BaseStream.Position;
                    Raw.Size = Raw.Raw.Length;
                    Raw.RawLocation = RawLocation.Internal;

                    //Store EOF
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                    //Write Padding
                    map.bw.BaseStream.Position = Raw.Offset;
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);

                    //Write Raw
                    Raw.Offset = (int)map.bw.BaseStream.Position;
                    map.bw.Write(Raw.Raw);

                    //WriteEOF
                    map.bw.Write(EOF);

                    //Add Shift
                    Shift += Raw.Size;
                    Shift += Padding;

                    //Shift Raw Table Size
                    map.Header.ModelRawTableSize += Shift;
                }
                #endregion

                #region Weather
                if (Tags[i].EndsWith(".weat"))
                {
                    //Set Raw Type
                    RawType = RawType.Weather;

                    //Goto the Last Weather Pointer
                    int Off = 0;
                    int Size = 0;
                    for (int x = 0; x < Weather.Count; x++)
                    {
                        map.br.BaseStream.Position = Weather[x].MetaOffset + 56;
                        int Offset = map.br.ReadInt32();
                        int Size2 = map.br.ReadInt32();
                        if (Offset > 0 && Size > 0 && RawData.RawLocator(Offset) == RawLocation.Internal)
                        {
                            if (Offset > Off)
                            {
                                Off = Offset;
                                Size = Size2;
                            }
                        }
                    }
                    map.br.BaseStream.Position = Off + Size;

                    //Open and Read Raw
                    Raw = new RawChunk();
                    Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                    Raw.Offset = (int)map.br.BaseStream.Position;
                    Raw.Size = Raw.Raw.Length;
                    Raw.RawLocation = RawLocation.Internal;

                    //Store EOF
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                    //Write Padding
                    map.bw.BaseStream.Position = Raw.Offset;
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);

                    //Write Raw
                    Raw.Offset = (int)map.bw.BaseStream.Position;
                    map.bw.Write(Raw.Raw);

                    //Write EOF
                    map.bw.Write(EOF);

                    //Add to Shift
                    Shift += Raw.Size;
                    Shift += Padding;

                    //Shift Raw Table Size
                    map.Header.ModelRawTableSize += Shift;
                }
                #endregion

                #region Particle Raw
                if (Tags[i].EndsWith(".PRTM"))
                {
                    //Set Raw Type
                    RawType = RawType.Particle;

                    //Goto the end of the model raw table
                    map.bw.BaseStream.Position = map.Header.ModelRawTableStart + map.Header.ModelRawTableSize;

                    //Open and Read Raw
                    Raw = new RawChunk();
                    Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                    Raw.Offset = (int)map.br.BaseStream.Position;
                    Raw.Size = Raw.Raw.Length;
                    Raw.RawLocation = RawLocation.Internal;

                    //Store EOF
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                    //Write Padding
                    map.bw.BaseStream.Position = Raw.Offset;
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);

                    //Write Raw
                    Raw.Offset = (int)map.bw.BaseStream.Position;
                    map.bw.Write(Raw.Raw);

                    //Write EOF
                    map.bw.Write(EOF);

                    //Add to Shift
                    Shift += Raw.Size;
                    Shift += Padding;

                    //Shift Raw Table Size
                    map.Header.ModelRawTableSize += Shift;
                }
                #endregion

                #region Read the new Tag
                //Create a new Tag
                Tag NewTag = new Tag();

                //Open the Xml
                XmlTextReader xr = new XmlTextReader(new FileStream(Dir + Tags[i], FileMode.Open, FileAccess.Read, FileShare.Read));
                while (xr.Read())
                {
                    #region Read Xml
                    switch (xr.NodeType)
                    {
                        case XmlNodeType.Element:
                            {
                                switch (xr.Name.ToLower())
                                {
                                    case "tag":
                                        {
                                            xr.ReadStartElement();
                                            NewTag.Class = xr.ReadElementString("Class");
                                            NewTag.Path = xr.ReadElementString("Path");
                                            xr.ReadElementString("Ident");
                                            NewTag.MetaSize = Convert.ToInt32(xr.ReadElementString("Size"));
                                            int IdentCount = Convert.ToInt32(xr.ReadElementString("IdentCount"));
                                            int ReflexiveCount = Convert.ToInt32(xr.ReadElementString("ReflexiveCount"));
                                            int ScriptCount = Convert.ToInt32(xr.ReadElementString("ScriptCount"));

                                            //Idents
                                            for (int x = 0; x < IdentCount; x++)
                                            {
                                                TagReference t = new TagReference();
                                                t.Tag = new Tag();
                                                xr.ReadStartElement();
                                                t.Tag.Class = xr.ReadElementString("Class");
                                                t.Tag.Path = xr.ReadElementString("Path");
                                                t.Tag.Identifier = Convert.ToInt32(xr.ReadElementString("Ident"));
                                                t.Offset = Convert.ToInt32(xr.ReadElementString("Offset"));
                                                xr.ReadEndElement();
                                                NewTag.Idents.Add(t);
                                            }

                                            //Reflexives
                                            for (int x = 0; x < ReflexiveCount; x++)
                                            {
                                                Reflexive r = new Reflexive();
                                                xr.ReadStartElement();
                                                r.ChunkCount = Convert.ToInt32(xr.ReadElementString("ChunkCount"));
                                                r.ChunkSize = Convert.ToInt32(xr.ReadElementString("ChunkSize"));
                                                r.Translation = Convert.ToInt32(xr.ReadElementString("Translation"));
                                                r.Offset = Convert.ToInt32(xr.ReadElementString("Offset"));
                                                xr.ReadEndElement();
                                                NewTag.Reflexives.Add(r);
                                            }

                                            //Scripts
                                            for (int x = 0; x < ScriptCount; x++)
                                            {
                                                ScriptReference s = new ScriptReference();
                                                xr.ReadStartElement();
                                                s.Offset = Convert.ToInt32(xr.ReadElementString("Offset"));
                                                s.Index = Convert.ToInt32(xr.ReadElementString("Index"));
                                                s.Reference = xr.ReadElementString("Reference");
                                                xr.ReadEndElement();
                                                NewTag.ScriptReferences.Add(s);
                                            }
                                            break;
                                        }
                                }
                                break;
                            }
                    }
                    #endregion
                }
                #endregion

                #region Write Internal Raw Pointers
                if (Tags[i].EndsWith(".snd!") | Tags[i].EndsWith(".mode"))
                {
                    goto Model;
                }
                else if (Tags[i].EndsWith(".sbsp"))
                {
                    goto BSP;
                }
                else if (Tags[i].EndsWith(".ltmp"))
                {
                    goto Lightmap;
                }
                else if (Tags[i].EndsWith(".DECR"))
                {
                    goto Decorator;
                }
                else if (Tags[i].EndsWith(".weat"))
                {
                    goto Weather;
                }
                else if (Tags[i].EndsWith(".PRTM"))
                {
                    goto Particle;
                }


            Model:
                {
                    foreach (Tag t in Models)
                    {
                        //Model 1
                        map.br.BaseStream.Position = t.MetaOffset + 36 + Shift;
                        Reflexive r = new Reflexive();
                        r.Read(map, true);
                        for (int x = 0; x < r.ChunkCount; x++)
                        {
                            map.br.BaseStream.Position = r.Translation + (x * 92) + 56 + Shift;
                            int Offset = map.br.ReadInt32();
                            if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Shift; }
                            map.bw.BaseStream.Position = r.Translation + (x * 92) + 56 + Shift;
                            map.bw.Write(Offset);
                        }

                        //Model 2
                        map.br.BaseStream.Position = t.MetaOffset + 116 + Shift;
                        r.Read(map, true);
                        for (int x = 0; x < r.ChunkCount; x++)
                        {
                            map.br.BaseStream.Position = r.Translation + (x * 88) + 52 + Shift;
                            int Offset = map.br.ReadInt32();
                            if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                            map.bw.BaseStream.Position = r.Translation + (x * 88) + 52 + Shift;
                            map.bw.Write(Offset);
                        }
                    }
                    goto BSP;
                }
            BSP:
                {
                    int Offset = 0;
                    foreach (Bsp bsp in map.Bsps)
                    {
                        //Detail Object
                        foreach (DetailObject d in bsp.DetailObjects)
                        {
                            map.br.BaseStream.Position = d.RawPointerOffset + Shift;
                            Offset = map.br.ReadInt32();
                            if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                            map.bw.BaseStream.Position = d.RawPointerOffset + Shift;
                            map.bw.Write(Offset);
                        }

                        //Permutations
                        foreach (BSPPermutation p in bsp.Permutations)
                        {
                            map.br.BaseStream.Position = p.RawPointerOffset + Shift;
                            Offset = map.br.ReadInt32();
                            if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                            map.bw.BaseStream.Position = p.RawPointerOffset + Shift;
                            map.bw.Write(Offset);
                        }

                        //Water Definitions
                        foreach (WaterDefinition w in bsp.WaterDefinitions)
                        {
                            map.br.BaseStream.Position = w.RawPointerOffset + Shift;
                            Offset = map.br.ReadInt32();
                            if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                            map.bw.BaseStream.Position = w.RawPointerOffset + Shift;
                            map.bw.Write(Offset);
                        }

                        //Decorator
                        foreach (BSP.Decorator d in bsp.Decorators)
                        {
                            foreach (CacheBlock c in d.CacheBlocks)
                            {
                                map.br.BaseStream.Position = c.RawPointerOffset + Shift;
                                Offset = map.br.ReadInt32();
                                if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                                map.bw.BaseStream.Position = c.RawPointerOffset + Shift;
                                map.bw.Write(Offset);
                            }
                        }
                    }
                    goto Lightmap;
                }
            Lightmap:
                {
                    int Offset = 0;
                    foreach (Lightmap l in map.Lightmaps)
                    {
                        //Lightmap Groups
                        foreach (LightmapGroup g in l.Groups)
                        {
                            //Clutors
                            foreach (Clustor c in g.Clustors)
                            {
                                map.br.BaseStream.Position = c.RawPointerOffset + Shift;
                                Offset = map.br.ReadInt32();
                                if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                                map.bw.BaseStream.Position = c.RawPointerOffset + Shift;
                                map.bw.Write(Offset);
                            }

                            //Poop
                            foreach (PoopDefinition p in g.PoopDefintions)
                            {
                                map.br.BaseStream.Position = p.RawPointerOffset + Shift;
                                Offset = map.br.ReadInt32();
                                if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                                map.bw.BaseStream.Position = p.RawPointerOffset + Shift;
                                map.bw.Write(Offset);
                            }

                            //GeometryBuckets
                            foreach (GeometryBucket b in g.GeometryBuckets)
                            {
                                map.br.BaseStream.Position = b.RawPointerOffset + Shift;
                                Offset = map.br.ReadInt32();
                                if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                                map.bw.BaseStream.Position = b.RawPointerOffset + Shift;
                                map.bw.Write(Offset);
                            }
                        }
                    }
                    goto Decorator;
                }
            Decorator:
                {
                    int Offset = 0;
                    foreach (Tag t in Decorators)
                    {
                        map.br.BaseStream.Position = t.MetaOffset + 56 + Shift;
                        Offset = map.br.ReadInt32();
                        if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                        map.bw.BaseStream.Position = t.MetaOffset + 56 + Shift;
                        map.bw.Write(Offset);
                    }
                    goto Weather;
                }
            Weather:
                {
                    int Offset = 0;
                    foreach (Tag t in Weather)
                    {
                        map.br.BaseStream.Position = t.MetaOffset + Shift;
                        Reflexive r = new Reflexive();
                        r.Read(map, true);
                        for (int x = 0; x < r.ChunkCount; x++)
                        {
                            map.br.BaseStream.Position = r.Translation + (x * 140) + 65 + Shift;
                            Offset = map.br.ReadInt32();
                            if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                            map.bw.BaseStream.Position = r.Translation + (x * 140) + 65 + Shift;
                            map.bw.Write(Offset);
                        }
                    }
                    goto Particle;
                }
            Particle:
                {
                    int Offset = 0;
                    foreach (Tag t in Particles)
                    {
                        map.br.BaseStream.Position = t.MetaOffset + 160 + Shift;
                        Offset = map.br.ReadInt32();
                        if (Offset > 0 && RawData.RawLocator(Offset) == RawLocation.Internal) { Offset += Raw.Size; }
                        map.bw.BaseStream.Position = t.MetaOffset + 160 + Shift;
                        map.bw.Write(Offset);
                    }
                }
                #endregion

                #region Add new StringIDs
                //Add new StringIds
                foreach (ScriptReference s in NewTag.ScriptReferences)
                {
                    if (map.FindScript(s.Reference) == 0)
                    {
                        ScriptReference NewS = new ScriptReference();
                        NewS.Reference = s.Reference;
                        NewS.Length = (byte)s.Reference.Length;
                        NewS.Index = map.Header.SIDCount;

                        //StringId 128
                        map.Header.SIDMetaTableOffset += Shift;
                        OldPadding = map.Header.SIDMetaTableOffset + (128 * map.Header.SIDCount) - map.Header.SIDIndexOffset;
                        map.br.BaseStream.Position = map.Header.SIDIndexOffset + Shift;
                        EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                        map.bw.BaseStream.Position = map.Header.SIDMetaTableOffset + (map.Header.SIDCount * 128);
                        map.bw.Write(NewS.Reference.ToCharArray());
                        for (int x = NewS.Reference.Length; x < 128; x++)
                        {
                            map.bw.Write((byte)0);
                        }
                        Padding = map.Padding();
                        map.bw.Write(new byte[Padding]);
                        map.bw.Write(EOF);
                        Shift += 128 + Padding;
                        Shift -= OldPadding;

                        //StringId Index
                        map.Header.SIDIndexOffset += Shift;
                        OldPadding = map.Header.SIDIndexOffset + (4 * map.Header.SIDCount) - map.Header.SIDTableOffset;
                        map.br.BaseStream.Position = map.Header.SIDTableOffset + Shift;
                        EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                        map.bw.BaseStream.Position = map.Header.SIDIndexOffset + (4 * map.Header.SIDCount);
                        map.bw.Write((int)map.Header.SIDCount * 4);
                        Padding = map.Padding();
                        map.bw.Write(new byte[Padding]);
                        map.bw.Write(EOF);
                        Shift += 4 + Padding;
                        Shift -= OldPadding;

                        //StringId Table
                        map.Header.SIDTableOffset += Shift;
                        OldPadding = map.Header.SIDTableOffset + map.Header.SIDTableSize - map.Header.FileTableIndexOffset;
                        map.br.BaseStream.Position = map.Header.FileTableOffset + Shift;
                        EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                        map.bw.BaseStream.Position = map.Header.SIDTableOffset + map.Header.SIDTableSize;
                        map.bw.Write((byte)0);
                        map.bw.Write(NewS.Reference.ToCharArray());
                        Padding = map.Padding();
                        map.bw.Write(new byte[Padding]);
                        map.bw.Write(EOF);
                        map.Header.SIDTableSize += NewS.Reference.Length + 1;
                        Shift += NewS.Reference.Length + 1 + Padding;
                        Shift -= OldPadding;
                        map.Header.SIDCount += 1;
                        map.ScriptReferences.Add(NewS);
                    }
                }
                #endregion

                #region FileTable
                //Goto the File Table
                map.Header.FileTableOffset += Shift;
                OldPadding = map.Header.FileTableIndexOffset - (map.Header.FileTableOffset + map.Header.FileTableSize) + Shift;
                map.br.BaseStream.Position = map.Header.FileTableIndexOffset + Shift;
                EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                map.bw.BaseStream.Position = Ugh.FileTableOffset + Shift;

                //Write New Path
                int num = 1;
                NewTag.Path += "(";
                string TempS = NewTag.Path + "1)";
                while (map.FindIdent(NewTag.Class, TempS) != -1)
                {
                    num++;
                    TempS = TempS.Remove(TempS.Length - 2);
                    TempS += num.ToString() + ")";
                }
                NewTag.Path = TempS;
                NewTag.FileTableOffset = (int)map.bw.BaseStream.Position;
                map.bw.Write(NewTag.Path.ToCharArray());
                map.bw.Write((byte)0);
                Ugh.FileTableOffset = (int)map.bw.BaseStream.Position;
                map.bw.Write(Ugh.Path.ToCharArray());
                map.bw.Write((byte)0);
                map.Header.FileTableSize += NewTag.Path.Length + 1;
                Padding = map.Padding();
                map.bw.Write(new byte[Padding]);
                map.bw.Write(EOF);
                Shift += NewTag.Path.Length + 1 + Padding;
                Shift -= OldPadding;

                //Goto the FileTableIndex
                map.Header.FileTableIndexOffset += Shift;
                OldPadding = map.Unicode.Tables[0].IndexOffset - (map.Header.FileTableIndexOffset + (map.Index.TagCount * 4)) + Shift;
                map.br.BaseStream.Position = map.Unicode.Tables[0].IndexOffset + Shift;
                EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                map.bw.BaseStream.Position = map.Header.FileTableIndexOffset + (map.Index.TagCount * 4) - 4;
                map.bw.Write(NewTag.FileTableOffset - map.Header.FileTableOffset);
                map.bw.Write((NewTag.FileTableOffset - map.Header.FileTableOffset) + Shift);
                Padding = map.Padding();
                map.bw.Write(new byte[Padding]);
                map.bw.Write(EOF);
                Shift += 4 + Padding;
                Shift -= OldPadding;

                //Shift Unicode Table Pointers
                map.br.BaseStream.Position = map.Tags[0].MetaOffset + 400 + Shift;
                for (int x = 0; x < 9; x++)
                {
                    map.br.BaseStream.Position += 8;
                    int Index = map.br.ReadInt32();
                    int Table = map.br.ReadInt32();
                    map.bw.BaseStream.Position -= 8;
                    map.bw.Write(Index + Shift);
                    map.bw.Write(Table + Shift);
                    map.br.BaseStream.Position += 12;
                }
                #endregion

                #region Bitmap Raw
                if (Tags[i].EndsWith(".bitm"))
                {
                    //Set Raw Type
                    RawType = RawType.Bitmap;

                    //Goto the Index Offset
                    map.br.BaseStream.Position = map.Header.IndexOffset;

                    //Open and Read Raw
                    Raw = new RawChunk();
                    Raw.Raw = File.ReadAllBytes(Dir + Tags[i].Replace(".rawxml", ".raw"));
                    Raw.Offset = (int)map.br.BaseStream.Position;
                    Raw.Size = Raw.Raw.Length;
                    Raw.RawLocation = RawLocation.Internal;

                    //Store EOF
                    EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));

                    //Write Raw
                    Raw.Offset = (int)map.bw.BaseStream.Position;
                    map.bw.Write(Raw.Raw);

                    //Write Padding
                    Padding = map.Padding();
                    map.bw.Write(new byte[Padding]);

                    //Write EOF
                    map.bw.Write(EOF);

                    //Add to Shift
                    Shift += Raw.Size;
                    Shift += Padding;
                }
                #endregion

                #region Write Internal Bitmap Pointers
                //Relink Internal Raw Pointers
                foreach (Tag t in Bitmaps)
                {
                    Reflexive BitmapData = new Reflexive();
                    BitmapData.Read(map, t.MetaOffset + 68 + Shift, true);
                    for (int Chunk = 0; Chunk < BitmapData.ChunkCount; Chunk++)
                    {
                        map.br.BaseStream.Position = BitmapData.Translation + (Chunk * 116) + 28 + Shift;
                        int Offset1 = map.br.ReadInt32();
                        int Offset2 = map.br.ReadInt32();
                        int Offset3 = map.br.ReadInt32();
                        if (Offset1 > 0 && RawData.RawLocator(Offset1) == RawLocation.Internal) 
                        { 
                            Offset1 += Shift;
                            map.bw.BaseStream.Position = BitmapData.Translation + (Chunk * 116) + 28 + Shift;
                            map.bw.Write(Offset1);
                        }
                        if (Offset2 > 0 && RawData.RawLocator(Offset2) == RawLocation.Internal) 
                        { 
                            Offset2 += Shift;
                            map.bw.BaseStream.Position = BitmapData.Translation + (Chunk * 116) + 28 + Shift + 4;
                            map.bw.Write(Offset2);
                        }
                        if (Offset3 > 0 && RawData.RawLocator(Offset3) == RawLocation.Internal) 
                        { 
                            Offset3 += Shift;
                            map.bw.BaseStream.Position = BitmapData.Translation + (Chunk * 116) + 28 + Shift + 8;
                            map.bw.Write(Offset3);
                        }
                    }
                }
                #endregion

                #region Index
                //Pad the Index
                map.br.BaseStream.Position = map.Header.IndexOffset + Shift;
                EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                map.br.BaseStream.Position = map.Header.IndexOffset + Shift - 1;
                OldPadding = 0;
                while (map.br.ReadByte() == 0)
                {
                    OldPadding += 1;
                    map.br.BaseStream.Position -= 2;
                }
                Padding = map.Padding();
                map.bw.Write(new byte[Padding]);
                map.bw.Write(EOF);
                Shift += Padding;
                Shift -= OldPadding;

                //Add it to the Index
                map.Header.IndexOffset += Shift;
                map.Index.TagStart += Shift;
                map.br.BaseStream.Position = map.Index.TagStart + (map.Header.TagCount * 16) - 16;
                EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                map.bw.BaseStream.Position = map.Index.TagStart + (map.Header.TagCount * 16) - 16;

                Shift += 16;
                map.Header.IndexSize += 16;
                int NewMagic = map.Index.SecondaryMgicConstant - (map.Header.IndexOffset + map.Header.IndexSize);
                map.bw.Write(Strings.StrReverse(NewTag.Class).ToCharArray());
                map.bw.Write((int)Ugh.Identifier);
                map.bw.Write((int)(Ugh.MetaOffset + Shift) + NewMagic);
                map.bw.Write((int)NewTag.MetaSize);
                int Place = (int)map.bw.BaseStream.Position;                

                //Write Ugh back
                map.bw.Write(EOF);
                map.bw.BaseStream.Position = Place + 4;
                map.bw.Write((int)Ugh.Identifier + 65537);
                map.bw.Write((int)(Ugh.MetaOffset + NewTag.MetaSize + Shift) + NewMagic);
                map.bw.Write((int)Ugh.MetaSize);

                //Add it to the Meta Table
                NewTag.MetaOffset = Ugh.MetaOffset;
                map.br.BaseStream.Position = NewTag.MetaOffset + Shift;
                EOF = map.br.ReadBytes((int)(map.br.BaseStream.Length - map.br.BaseStream.Position));
                map.bw.BaseStream.Position = NewTag.MetaOffset + Shift;
                map.bw.Write(File.ReadAllBytes(Dir + Tags[i].Remove(Tags[i].LastIndexOf('.'))));
                map.bw.Write(EOF);
                map.Header.TagCount += 1;
                map.Index.TagCount += 1;
                map.Header.MetaTableSize += NewTag.MetaSize;
                map.Header.NonRawSize += NewTag.MetaSize + 16;
                //Shift += NewTag.MetaSize;
                #endregion

                #region Relink References
                //Relink all reflexives
                foreach (Reflexive r in NewTag.Reflexives)
                {
                    if (r.Translation > 0)
                    {
                        map.bw.BaseStream.Position = Ugh.MetaOffset + r.Offset + 4 + Shift;
                        map.bw.Write((Ugh.MetaOffset + r.Translation + Shift) + NewMagic);
                    }
                    else
                    {
                        map.bw.BaseStream.Position = Ugh.MetaOffset + r.Offset + Shift;
                        map.bw.Write((int)0);
                        map.bw.Write((int)-1 + NewMagic);
                    }
                }

                //Relink All Idents
                foreach (TagReference t in NewTag.Idents)
                {
                    map.bw.BaseStream.Position = NewTag.MetaOffset + t.Offset + 4;
                    int Index = Tags.IndexOf(t.Tag.Path);
                    if (Index != -1)
                    {
                        map.bw.Write(Idents[Index]);
                    }
                }

                //Relink all StringIDs
                foreach (ScriptReference s in NewTag.ScriptReferences)
                {
                    map.bw.BaseStream.Position = NewTag.MetaOffset + s.Offset;
                    map.bw.Write((short)map.FindScript(s.Reference));
                    map.bw.Write((byte)0);
                    map.bw.Write((byte)s.Reference.Length);
                }
                #endregion

                #region Fix Crazy Reflexives
                Reflexive re = new Reflexive();
                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 8 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 16 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 24 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 32 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 40 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 48 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);
                for (int x = 0; x < re.ChunkCount; x++)
                {
                    Reflexive rr = new Reflexive();
                    map.br.BaseStream.Position = re.Translation + (x * 52) + 20 + Shift;
                    rr.Read(map, true);
                    rr.Translation += NewTag.MetaSize;
                    re.Write(map, true);
                }

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 56 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 64 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 72 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);
                for (int x = 0; x < re.ChunkCount; x++)
                {
                    Reflexive rr = new Reflexive();
                    map.br.BaseStream.Position = re.Translation + (x * 28) + Shift;
                    rr.Read(map, true);
                    rr.Translation += NewTag.MetaSize;
                    re.Write(map, true);

                    map.br.BaseStream.Position = re.Translation + (x * 28) + 8 + Shift;
                    rr.Read(map, true);
                    rr.Translation += NewTag.MetaSize;
                    re.Write(map, true);
                }

                map.br.BaseStream.Position = NewTag.MetaOffset + NewTag.MetaSize + 80 + Shift;
                re.Read(map, true);
                re.Translation += NewTag.MetaSize;
                re.Write(map, true);
                for (int x = 0; x < re.ChunkCount; x++)
                {
                    Reflexive rr = new Reflexive();
                    map.br.BaseStream.Position = re.Translation + (x * 44) + 24 + Shift;
                    rr.Read(map, true);
                    rr.Translation += NewTag.MetaSize;
                    re.Write(map, true);
                }
                #endregion

                #region Fix Crazy LoneIds
                for (int x = Ugh.MetaOffset + NewTag.MetaSize + Shift; x < Ugh.MetaOffset + NewTag.MetaSize + Shift + Ugh.MetaSize; x += 4)
                {
                    map.br.BaseStream.Position = x;
                    int Temp = map.br.ReadInt32();
                    if (Temp == Ugh.Identifier)
                    {
                        map.bw.BaseStream.Position = x;
                        map.bw.Write((int)Ugh.Identifier + 65537);
                    }
                }
                #endregion

                #region Refresh
                map.TagHandler.Read(map);
                #endregion
            }

            //Update
            map.bw.BaseStream.Position = 8;
            map.bw.Write((int)map.bw.BaseStream.Length);
            map.bw.BaseStream.Position += 4;
            map.bw.Write(map.Header.IndexOffset);
            map.bw.Write(map.Header.IndexSize);
            map.bw.Write(map.Header.MetaTableSize);
            map.bw.Write(map.Header.NonRawSize);
            map.bw.BaseStream.Position = 364;
            map.bw.Write(map.Header.SIDMetaTableOffset);
            map.bw.Write(map.Header.SIDCount);
            map.bw.Write(map.Header.SIDTableSize);
            map.bw.Write(map.Header.SIDIndexOffset);
            map.bw.Write(map.Header.SIDTableOffset);
            map.bw.BaseStream.Position = 716;
            map.bw.Write(map.Header.TagCount);
            map.bw.Write(map.Header.FileTableOffset);
            map.bw.Write(map.Header.FileTableSize);
            map.bw.Write(map.Header.FileTableIndexOffset);
            map.bw.BaseStream.Position = 744;
            map.bw.Write(map.Header.ModelRawTableStart);
            map.bw.Write(map.Header.ModelRawTableSize);
            map.bw.BaseStream.Position = map.Header.IndexOffset + 24;
            map.bw.Write(map.Index.TagCount);
        }
    }
}
