﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace HaloMap
{
    public class MapSlot
    {
        #region Fields
        public int CampaignId;
        public int MapId;
        public int Start;
        public TagReference Bitm = new TagReference();
        public Image Preview;
        public string[] Name = new string[8];
        public string[] Description = new string[8];
        public string Scenario;
        public int SortOrder;
        public short Flags;
        public byte[] MaxTeams = new byte[15];
        #endregion

        #region Methods
        public void ReadMultiplayer(Map map)
        {
            MapId = map.br.ReadInt32();
            Bitm.Read(map, true);
            map.br.BaseStream.Position += 512;
            for (int i = 0; i < 32; i++)
            {
                Name[0] += map.br.ReadChar();
                map.br.BaseStream.Position++;
            }
            for (int i = 0; i < 128; i++)
            {
                Description[0] += map.br.ReadChar();
                map.br.BaseStream.Position++;
            }
            map.br.BaseStream.Position += 256;
            SortOrder = map.br.ReadInt32();
            Flags = map.br.ReadInt16();
            for (int i = 0; i < 15; i++)
            {
                MaxTeams[i] = map.br.ReadByte();
            }

            if (Bitm.Tag.Identifier != -1)
            {
                Tag Old = map.SelectedTag;
                map.SelectedTag = Bitm.Tag;
                BitmapRaw.Read(map);
                Preview = Bitmap.LoadPreview(map, BitmapRaw.RawChunks[0]);
                map.SelectedTag = Old;
            }
        }

        public void ReadCampaign(Map map)
        {
            MapId = map.br.ReadInt32();
            CampaignId = map.br.ReadInt32();
            Bitm.Read(map, true);
            map.br.BaseStream.Position += 512;
            for (int i = 0; i < 32; i++)
            {
                Name[0] += map.br.ReadChar();
                map.br.BaseStream.Position++;
            }
            for (int i = 0; i < 128; i++)
            {
                Description[0] += map.br.ReadChar();
                map.br.BaseStream.Position++;
            }

            if (Bitm.Tag.Identifier != -1)
            {
                Tag Old = map.SelectedTag;
                map.SelectedTag = Bitm.Tag;
                BitmapRaw.Read(map);
                Preview = Bitmap.LoadPreview(map, BitmapRaw.RawChunks[0]);
                map.SelectedTag = Old;
            }
        }
        #endregion
    }

    public class MainMenu
    {
        #region Fields
        public static Reflexive LevelData = new Reflexive();
        public static Reflexive Campaign = new Reflexive();
        public static List<MapSlot> SPLevels = new List<MapSlot>();
        public static Reflexive Multiplayer = new Reflexive();
        public static List<MapSlot> MPLevels = new List<MapSlot>();
        #endregion

        #region Methods
        public static void Read(Map map)
        {
            SPLevels.Clear();
            MPLevels.Clear();
            map.br.BaseStream.Position = map.Tags[0].MetaOffset + 376;
            LevelData.Read(map, true);
            map.br.BaseStream.Position = LevelData.Translation + 8;
            Campaign.Read(map, true);
            for (int i = 0; i < Campaign.ChunkCount; i++)
            {
                map.br.BaseStream.Position = Campaign.Translation + (i * 2896);
                MapSlot sp = new MapSlot();
                sp.ReadCampaign(map);
                SPLevels.Add(sp);
            }
            map.br.BaseStream.Position = LevelData.Translation + 16;
            Multiplayer.Read(map, true);
            for (int i = 0; i < Multiplayer.ChunkCount; i++)
            {
                map.br.BaseStream.Position = Multiplayer.Translation + (i * 3172);
                MapSlot mp = new MapSlot();
                mp.ReadMultiplayer(map);
                MPLevels.Add(mp);
            }
        }
        #endregion
    }
}
