﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Windows.Forms;
//using Microsoft.DirectX.DirectSound;

namespace HaloMap
{
    public enum Compression
    {
        None,
        XboxAdpmc,
        ImaAdpmc,
        Ogg,
        WMA
    }

    public class Sound
    {
        public Compression Format;
        public MemoryStream snd;
        public byte[] Data;

        public Sound(Map map)
        {
            foreach (SoundChoice cc in SoundRaw.Permutation.Choices)
            {
                foreach (SoundChunk c in cc.SoundChunks)
                {
                    if (c.RawLocation == RawLocation.Internal && c.Offset != -1 && c.Size > 0)
                    {
                        map.br.BaseStream.Position = c.Offset;
                        c.Raw = map.br.ReadBytes(c.Size);
                    }
                    else if (c.RawLocation != RawLocation.Internal && c.Offset != -1 && c.Size > 0)
                    {
                        BinaryReader br = map.OpenMap(c.RawLocation);
                        br.BaseStream.Position = c.Offset;
                        c.Raw = br.ReadBytes(c.Size);
                        br.Close();
                    }
                }
            }
        }

        public void FormatFile(Map map, SoundChunk Chunk)
        {
            snd = new MemoryStream();
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 4;
            byte format = map.br.ReadByte();
            switch (format)
            {
                case 0:
                case 1:
                    {//Wav
                        WriteADPMC(format, Chunk.Raw, Application.StartupPath + "\\TempSound.wav");
                        break;
                    }
                //case 2:
                //    {//Wma
                //        snd.Write(Chunk.Raw, 0, Chunk.Raw.Length);
                //        break;
                //    }
            }
        }

        public void FormatFile(Map map, SoundChunk[] Chunks)
        {
            snd = new MemoryStream();
            List<byte> Raw = new List<byte>();
            for (int i = 0; i < Chunks.Length; i++)
            {
                Raw.AddRange(Chunks[i].Raw);
            }
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 4;
            byte format = map.br.ReadByte();
            switch (format)
            {
                case 0:
                case 1:
                    {//Wav
                        WriteADPMC(format, Raw.ToArray(), Application.StartupPath + "\\TempSound.wav");
                        break;
                    }
                //case 2:
                //    {//Wma
                //        snd.Write(Chunk.Raw, 0, Chunk.Raw.Length);
                //        break;
                //    }
            }
        }

        public string LoadPreview(Map map, SoundChunk Chunk)
        {
            string Path = "";
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 4;
            int format = map.br.ReadByte();
            if (format == 0 | format == 1)
            {
                FormatFile(map, Chunk);
                Path = Application.StartupPath + "\\TempSound.wav";
            }
            //else if (format == 2)
            //{
            //    FormatFile(map, Chunk);
            //    Path = Application.StartupPath + "TempSound.wma";
            //}
            return Path;
        }

        public string PreviewAll(Map map, SoundChunk[] Chunk)
        {
            string Path = "";
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 4;
            int format = map.br.ReadByte();
            if (format == 0 | format == 1)
            {
                FormatFile(map, Chunk);
                Path = Application.StartupPath + "\\TempSound.wav";
            }
            //else if (format == 2)
            //{
            //    FormatFile(map, Chunk);
            //    Path = Application.StartupPath + "TempSound.wma";
            //}
            return Path;
        }

        public void Extract(string Location, Map map, SoundChunk Chunk)
        {
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 4;
            byte format = map.br.ReadByte();
            switch (format)
            {
                case 0:
                case 1:
                    {//Wav
                        WriteADPMC(format, Chunk.Raw, Location + ".wav");
                        break;
                    }
            }
        }

        public void ExtractAll(string Location, Map map, SoundChunk[] Chunks)
        {
            List<byte> Raw = new List<byte>();
            for (int i = 0; i < Chunks.Length; i++)
            {
                Raw.AddRange(Chunks[i].Raw);
            }
            map.br.BaseStream.Position = map.SelectedTag.MetaOffset + 4;
            byte format = map.br.ReadByte();
            switch (format)
            {
                case 0:
                case 1:
                    {//Wav
                        WriteADPMC(format, Raw.ToArray(), Location);
                        break;
                    }
            }
        }

        public void WriteADPMC(byte Format, byte[] Data, string Location)
        {
            int Channels = 0;
            int SampleRate = 0;
            switch (Format)
            {
                case 0:
                    {
                        Channels = 1;
                        SampleRate = 0x5622;
                        break;
                    }
                case 1:
                    {
                        Channels = 2;
                        SampleRate = 0xAC44;
                        break;
                    }
            }

            byte[] RIFF = { 52, 49, 46, 46 };
            byte[] WAVE = { 57, 41, 56, 45 };
            byte[] fmt = { 66, 0x6D, 74, 20 };
            byte[] data = { 64, 61, 74, 61 };

            //Write
            BinaryWriter bw = new BinaryWriter(new FileStream(Location, FileMode.Create, FileAccess.Write, FileShare.Write));
            bw.Write(("RIFF").ToCharArray());
            bw.Write(Data.Length + 44);
            bw.Write(("WAVE").ToCharArray());
            bw.Write(("fmt ").ToCharArray());
            bw.Write(0x14);
            bw.Write((ushort)0x69);
            bw.Write((ushort)Channels);
            bw.Write(SampleRate);
            bw.Write(Channels * SampleRate * 4 / 8);
            bw.Write((ushort)(Channels * 36));
            bw.Write((ushort)0x4);
            bw.Write((ushort)0x2);
            bw.Write((ushort)0x40);
            bw.Write(("data").ToCharArray());
            bw.Write(Data.Length);
            bw.Write(Data);
            bw.Close();
        }
    }
}
